// runoutput ./rotate.go

// Copyright 2013 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Generate test of bit rotations.
// The output is compiled and run.

package main

const mode = 0
