// Copyright 2017 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package main

type S struct{ i int }
type SS = S

func sub()

func main() {
	sub()
}
